#!/usr/bin/env python
"""
gisht

* This program is free software, see LICENSE file for details. *
"""
__version__ = "0.3"
__description__ = "Gists in the shell"
__author__ = "Karol Kuczmarski"
__license__ = "GPLv3"


import logging
import os
from pathlib import Path


#: Main application's directory.
APP_DIR = Path(os.path.expanduser('~/.gisht'))

#: Directory where gist repos are stored.
#: Subdirectories have names corresponding to numerical IDs of the gists.
GISTS_DIR = APP_DIR / 'gists'

#: Directory where links to gist "binaries" are stored.
#: Subdirectories have names corresponding to GitHub user handles
#: and contain symbolic links to executable files inside gist repos.
BIN_DIR = APP_DIR / 'bin'

#: Directory where the request cache resides.
#: Subdirectories contiain pickled :class:`Response` objects
#: (from :module:`requests`) inside directory paths corresponding to URL paths.
CACHE_DIR = APP_DIR / 'cache'


#: Logger object used by the application.
logger = logging.getLogger(__name__)



"""
Module implementing requests to GitHub API.
"""
from collections import OrderedDict
from datetime import datetime, timedelta
from pathlib import Path
import pickle

from hammock import Hammock
import requests

from gisht import CACHE_DIR
from gisht.util import ensure_path


__all__ = ['get_gist_info', 'iter_gists']


def get_gist_info(gist_id):
    """Retrieve information about gist of given ID.

    :param gist_id: Numerical ID of a GitHub gist
                    (NOT the user-visible <owner>/<name> string!)

    :return: Dictionary with gist information
    :raises: :class:`requests.exception.HTTPError`
    """
    if '/' in gist_id:
        raise ValueError("expected gist ID, not the <owner>/<name> reference!")

    github = GitHub()
    response = github.gists.GET(gist_id)
    response.raise_for_status()
    return response.json()


def iter_gists(owner):
    """Iterate over gists owned by given user.

    :param owner: GitHub user's name
    :return: Iterable (generator) of parsed JSON objects (dictionaries)
    :raises: :class:`requests.exception.HTTPError`
    """
    if type(owner).__name__ not in ('str', 'unicode'):
        raise TypeError("expected a string")

    def generator():
        github = GitHub()
        gists_url = str(github.users(owner).gists)
        while gists_url:
            gists_response = requests.get(
                gists_url, params={'per_page': GitHub.RESPONSE_PAGE_SIZE})
            gists_response.raise_for_status()

            for gist_json in to_json(gists_response):
                yield gist_json

            gists_url = gists_response.links.get('next', {}).get('url')

    return generator()


# API client

class CachedHammock(Hammock):
    """Custom version of :class:`Hammock` REST client that supports
    caching of GET requests.
    """
    def __init__(self, *args, **kwargs):
        self._cache_dir = kwargs.pop('cache_dir', None)
        self._cache_ttl = kwargs.pop('cache_ttl', timedelta(days=7))
        super(CachedHammock, self).__init__(*args, **kwargs)

    def _path(self):
        """Return only the path portion of the URL."""
        result = "/" if self._append_slash else ""
        path_comp = self
        while path_comp._parent:
            result = "/" + path_comp._name + result
            path_comp = path_comp._parent
        return result[1:] if len(result) > 1 else result  # no leading slash

    def _request(self, method, *args, **kwargs):
        """Make the HTTP request using :module:`requests` module."""
        use_cache = (method.upper() == 'GET' and
                     bool(self._cache_dir) and
                     kwargs.pop('cache', True))

        # try to load the response from cache, if available
        if use_cache:
            url_path = self._chain(*args)._path()
            cache_file = Path(self._cache_dir) / url_path
            if not self._expired(cache_file):
                with cache_file.open('rb') as f:
                    return pickle.load(f)

        # TODO(xion): if request has failed due to transient error,
        # return the cached response even if it's expired, and show appropriate
        # warning on stderr that data might be stale (on -i/--info)
        response = super(CachedHammock, self)._request(method, *args, **kwargs)
        # TODO(xion): the above logic should be also influenced by -/--local

        # save the obtained response to cache
        if use_cache:
            ensure_path(cache_file.parent)
            with cache_file.open('wb') as f:
                pickle.dump(response, f)

        return response

    def _expired(self, cache_file):
        """Check whether the cached response has expired."""
        if not cache_file.exists():
            return True
        last_modify_timestamp = cache_file.stat().st_mtime
        last_modify_time = datetime.fromtimestamp(last_modify_timestamp)
        return last_modify_time + self._cache_ttl <= datetime.now()


class GitHub(CachedHammock):
    """Client for GitHub REST API."""
    API_URL = 'https://api.github.com'

    #: Size of the GitHub response page in items (e.g. gists).
    RESPONSE_PAGE_SIZE = 50

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('cache_dir', CACHE_DIR / 'github')
        super(GitHub, self).__init__(self.API_URL, *args, **kwargs)


# Utility functions

def to_json(response):
    """Interpret given Requests' response object as JSON."""
    return response.json(object_pairs_hook=OrderedDict)



"""
Module for parsing command line arguments.
"""
import argparse
from enum import Enum
from itertools import chain
import logging

import argcomplete
import requests

from gisht import __version__, BIN_DIR
from gisht.github import iter_gists


__all__ = ['parse_argv', 'GistAction']


def parse_argv(argv):
    """Parse command line arguments.

    :param argv: List of command line argument strings,
                 *including* the program name in argv[0]

    :return: Parse result from :func:`argparse.ArgumentParser.parse_args`
    """
    # treat everything after -- in the command line as arguments to be passed
    # to the gist executable itself (rather than be parsed by us)
    # TODO(xion): implement this as custom argparse action with nargs=REMAINDER
    gist_args = []
    if '--' in argv:
        double_dash_pos = len(argv) - 1 - argv[::-1].index('--')  # last occur.
        gist_args = argv[double_dash_pos + 1:]
        argv = argv[:double_dash_pos]

    parser = create_argv_parser()

    # trigger autocompletion, unless gist args have been provided after --
    # (otherwise they may be considered as part of GIST argument)
    # TODO(xion): make ArgumentParser handle GIST_ARGS and remove this hack
    if not gist_args:
        argcomplete.autocomplete(parser)

    # TODO(xion): support reading default parameter values from ~/.gishtrc
    result = parser.parse_args(argv[1:])
    result.gist_args = gist_args
    return result


class GistAction(Enum):
    """Action to undertake towards the gist."""
    RUN = 'run'
    WHICH = 'which'
    PRINT = 'print'
    OPEN = 'open'
    INFO = 'info'


# Argument parser definition

def create_argv_parser():
    """Create a :class:`argparse.ArgumentParser` object
    for parsing command line arguments passed by the user.
    """
    parser = argparse.ArgumentParser(
        description="Download & run GitHub gists straight from the terminal",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        add_help=False)

    add_gist_group(parser)
    add_gist_action_group(parser)
    add_logging_group(parser)

    misc_group = add_misc_group(parser)

    # get the autogenerated usage string and tweak it a little
    # to include gist arguments that are handled separately
    # and exclude the miscellaneous flags which aren't part of a normal usage
    usage = parser.format_usage()
    usage = usage[usage.find(parser.prog):].rstrip("\n")  # remove cruft
    for misc_flag in chain.from_iterable(a.option_strings
                                         for a in misc_group._group_actions):
        usage = usage.replace(" [%s]" % misc_flag, "")
    parser.usage = usage + " [-- GIST_ARGS]"
    return parser


def add_gist_group(parser):
    """Include an argument group that allows to specify the gist.
    :param parser: :class:`argparse.ArgumentParser`
    :return: Resulting argument group
    """
    group = parser.add_argument_group(
        "Gist", "Specifies the gist, optionally with flags")

    group.add_argument('gist', type=gist,
                       help="GitHub gist, specified as <owner>/<name> "
                            "(e.g. Octocat/foo)",
                       metavar="GIST").completer = gist_completer

    group.add_argument('-l', '--local', '--cached',
                       default=False, action='store_true',
                       help="only run the gist if it's available locally "
                            "(do not fetch it from GitHub)")
    # TODO(xion): add a command line flag to always fetch the gist
    # (removing the existing one if necessary, or doing a `git pull`)

    return group


def add_gist_action_group(parser):
    """Include an argument group that allows to specify
    the action to be performed on the gist.

    :param parser: :class:`argparse.ArgumentParser`
    :return: Resulting argument group
    """
    group = parser.add_argument_group(
        "Actions", "Possible actions to perform on the gist") \
        .add_mutually_exclusive_group()
    group.set_defaults(action=GistAction.RUN)

    # TODO(xion): nargs='?' is illegal here for some inane reason, so these
    # flags can be specified more than once (if they're identical);
    # find a way to make that an error (probably another custom action -_-)
    group.add_argument('-r', '--run', dest='action',
                       action='store_const', const=GistAction.RUN,
                       help="run specified gist; this is the default behavior "
                            "if no action was specified explicitly")
    group.add_argument('-w', '--which', dest='action',
                       action='store_const', const=GistAction.WHICH,
                       help="output the path to binary which would be "
                            "ran for given gist; useful for passing it "
                            "to other commands via $( )")
    group.add_argument('-p', '--print', dest='action',
                       action='store_const', const=GistAction.PRINT,
                       help="print gist source on the standard output")
    group.add_argument('-o', '--open', dest='action',
                       action='store_const', const=GistAction.OPEN,
                       help="open the gist's GitHub page "
                            "in the default web browser")
    group.add_argument('-i', '--info', dest='action',
                       action='store_const', const=GistAction.INFO,
                       help="show summary information about specified gist")

    return group


def add_logging_group(parser):
    """Include an argument group that allows to
    control verbosity of the program's logging output.

    :param parser: :class:`argparse.ArgumentParser`
    :return: Resulting argument group
    """
    group = parser.add_argument_group(
        "Verbosity", "Only errors and warnings are printed by default") \
        .add_mutually_exclusive_group()
    group.set_defaults(log_level=LogLevelAction.DEFAULT_LEVEL)

    group.add_argument('-v', '--verbose', dest='log_level',
                       action=LogLevelAction,
                       const=-LogLevelAction.DEFAULT_INCREMENT,
                       help="include finer grained details in the output; "
                            "this option can be specified multiple times")
    group.add_argument('-q', '--quiet', dest='log_level',
                       action=LogLevelAction,
                       const=LogLevelAction.DEFAULT_INCREMENT,
                       help="decrease the verbosity level")

    return group


def add_misc_group(parser):
    """Include the argument group with miscellaneous options.
    :param parser: :class:`argparse.ArgumentParser`
    :return: Resulting argument group
    """
    group = parser.add_argument_group("Miscellaneous", "Other options")

    group.add_argument('--version', action='version', version=__version__)
    group.add_argument('-h', '--help', action='help',
                       help="show this help message and exit")

    return group


# Supporting code

def gist(value):
    """Converter/validator for the GIST command line argument."""
    try:
        owner, gist_name = value.split('/')
    except ValueError:
        raise argparse.ArgumentTypeError(
            "%r is not a valid gist reference; try '<owner>/`<name>`" % value)
    if owner and gist_name:
        return value
    else:
        raise argparse.ArgumentTypeError(
            "neither gist owner or name can be empty (got %r)" % (value,))


class LogLevelAction(argparse.Action):
    """Custom argument parser's :class:`Action` for handling
    log level / verbosity flags.
    """
    DEFAULT_LEVEL = logging.WARNING

    DEFAULT_INCREMENT = 10
    DEFAULT_MINIMUM = logging.NOTSET
    DEFAULT_MAXIMUM = logging.CRITICAL

    def __init__(self, *args, **kwargs):
        self.min = kwargs.pop('min', self.DEFAULT_MINIMUM)
        self.max = kwargs.pop('max', self.DEFAULT_MAXIMUM)
        kwargs.setdefault('const', self.DEFAULT_INCREMENT)
        super(LogLevelAction, self).__init__(*args, nargs=0, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        current = getattr(namespace, self.dest, self.DEFAULT_LEVEL)
        new = max(self.min, min(self.max, current + self.const))
        setattr(namespace, self.dest, new)


def gist_completer(prefix, parsed_args, **kwargs):
    """Autocompleter for the GIST command line argument.

    It tries to complete the gist identifier (<owner>/<name>) by listing
    all the gists for the typed <owner>, possibly filtered
    through a prefix of <name>.

    :return: Iterable of possible completions
    """
    results = set()

    # start with the locally available gists, possibly including entries
    # for GitHub users whose gists we have cached (if autocomplete prefix
    # does not include a slash)
    for entry in BIN_DIR.rglob('*'):
        entry = '/'.join(entry.relative_to(BIN_DIR).parts)
        if not entry.startswith(prefix):
            continue
        if '/' not in entry:
            entry = entry + '/'
        results.add(entry)
    # TODO(xion): the above is somewhat redundant in typical case, when GitHub
    # is available and user typed the owner part fully; elide it in this case

    # if username was given in full, query GitHub for that user's gist
    # and include them in the suggestions
    local = getattr(parsed_args, 'local', False)
    if '/' in prefix and not local:
        owner, name_prefix = prefix.split('/')
        try:
            for gist_json in iter_gists(owner):  # TODO(xion): cache this LOL
                for filename in gist_json['files'].keys():
                    if filename.startswith(name_prefix):
                        results.add(owner + '/' + filename)
                        break  # only add one (first) file per gist
        except requests.exceptions.HTTPError:
            pass  # limit to local completions if GitHub is unreachable

    return sorted(results)



"""
Module with utility functions used throughout the code.
"""
import logging
import os
from pathlib import Path
import sys

from gisht import logger


__all__ = [
    'ensure_path',
    'error', 'fatal',
]


def ensure_path(path):
    """Ensures given path exists, creating all necessary directory levels.
    Does nothing otherwise.
    """
    path = Path(path)
    if not path.exists():
        path.mkdir(parents=True)


# Error handling

def error(msg, *args, **kwargs):
    """Output an error message and end the program.
    :param exitcode: Optional keyword argument to specify the exit code
    """
    _exit(logging.ERROR, msg, *args, **kwargs)


def fatal(msg, *args, **kwargs):
    """Output an error message and end the program.
    :param exitcode: Optional keyword argument to specify the exit code
    """
    _exit(logging.CRITICAL, msg, *args, **kwargs)


def _exit(severity, msg, *args, **kwargs):
    """Log a message with given severity and exit the program."""
    msg = msg % args if args else msg
    logger.log(severity, "%s: error: %s", os.path.basename(sys.argv[0]), msg)
    raise SystemExit(kwargs.get('exitcode', 1))



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
"""
Executable script.
"""
from __future__ import print_function

import logging
import os
import sys

import requests

from gisht import APP_DIR, logger
from gisht.args import GistAction, parse_argv
from gisht.gists import (download_gist, gist_exists,
                         open_gist_page, output_gist_binary_path,
                         print_gist, run_gist, show_gist_info)
from gisht.util import error


def main(argv=sys.argv):
    """Entry point."""
    if os.name != 'posix':
        print("only POSIX operating systems are supported", file=sys.stderr)
        return os.EX_UNAVAILABLE

    args = parse_argv(argv)
    setup_logging(args.log_level)

    # during the first run, display a warning about executing untrusted code
    if not APP_DIR.exists():
        if logger.getEffectiveLevel() >= logging.WARNING:
            acknowledged = display_warning()
            if not acknowledged:
                return 2
        APP_DIR.mkdir(parents=True)

    gist = args.gist
    gist_args = args.gist_args

    # if the gist hasn't been cached locally, download it from GitHub
    if gist_exists(gist):
        logger.debug("gist %s found among already downloaded gists", gist)
    else:
        if args.local:
            error("gist %s is not available locally", gist,
                  exitcode=os.EX_NOINPUT)
        try:
            if not download_gist(gist):
                error("gist %s not found", gist, exitcode=os.EX_DATAERR)
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404:
                error("user '%s' not found", gist.split('/')[0],
                      exitcode=os.EX_UNAVAILABLE)
            else:
                error("HTTP error: %s", e, exitcode=os.EX_UNAVAILABLE)

    # do with the gist what the user has requested (default: run it)
    if args.action == GistAction.RUN:
        run_gist(gist, gist_args)
    elif args.action not in GistAction:
        error("unknown gist action %r", args.action, exitcode=os.EX_USAGE)
    else:
        if gist_args:
            error("gist arguments are only allowed when running the gist",
                  exitcode=os.EX_USAGE)
        if args.action == GistAction.WHICH:
            output_gist_binary_path(gist)
        elif args.action == GistAction.PRINT:
            print_gist(gist)
        elif args.action == GistAction.OPEN:
            open_gist_page(gist)
        elif args.action == GistAction.INFO:
            show_gist_info(gist)


def setup_logging(level):
    """Sets up the application-wide logging system to output everything to
    standard output, and filter our own messages according to given
    minimum level.
    """
    logger.setLevel(level)

    # propagate log messages from our application-specific logger
    # to the root logger, to be dumped to stderr alongside any possible output
    # from other loggers (like those from third party libraries)
    logger.propagate = True
    handler = logging.StreamHandler(sys.stderr)
    logging.getLogger().addHandler(handler)


def display_warning():
    """Displays a warning about executing untrusted code
    and ask the user to continue.
    :return: Whether the user chose to continue
    """
    print(
        "WARNING: gisht is used to download & run code from a remote source.",
        "",
        "Never run gists that you haven't authored, and/or do not trust.",
        "Doing so is dangerous, and may expose your system to security risks!",
        "",
        "(If you continue, this warning won't be shown again).",
        "",
        sep=os.linesep, file=sys.stderr)

    print("Do you want to continue? [y/N]: ", end="", file=sys.stderr)
    answer = raw_input()
    return answer.lower().strip() == 'y'


if __name__ == '__main__':
    sys.exit(main() or 0)



#!/usr/bin/env python
"""
gisht
=====

With *gisht*, you can run scripts published as GitHub gists
with a single command::

    gisht Xion/git-today

Behind the scenes, *gisht* will fetch the gist, cache it locally,
and run its code.

Refer to project's website for more information.
"""
import ast
import os
from setuptools import find_packages, setup
import sys


# Utility functions

def read_tags(filename):
    """Reads values of "magic tags" defined in the given Python file.

    :param filename: Python filename to read the tags from
    :return: Dictionary of tags
    """
    with open(filename) as f:
        ast_tree = ast.parse(f.read(), filename)

    res = {}
    for node in ast.walk(ast_tree):
        if type(node) is not ast.Assign:
            continue

        target = node.targets[0]
        if type(target) is not ast.Name:
            continue

        if not (target.id.startswith('__') and target.id.endswith('__')):
            continue

        name = target.id[2:-2]
        res[name] = ast.literal_eval(node.value)

    return res


def read_requirements(filename='requirements.txt'):
    """Reads the list of requirements from given file.

    :param filename: Filename to read the requirements from.
                     Uses ``'requirements.txt'`` by default.

    :return: Requirements as list of strings
    """
    # allow for some leeway with the argument
    if not filename.startswith('requirements'):
        filename = 'requirements-' + filename
    if not os.path.splitext(filename)[1]:
        filename += '.txt'  # no extension, add default

    def valid_line(line):
        line = line.strip()
        return line and not any(line.startswith(p) for p in ('#', '-'))

    def extract_requirement(line):
        egg_eq = '#egg='
        if egg_eq in line:
            _, requirement = line.split(egg_eq, 1)
            return requirement
        return line

    with open(filename) as f:
        lines = f.readlines()
        return list(map(extract_requirement, filter(valid_line, lines)))


# setup() call

install_requires = read_requirements()
if sys.version_info < (3, 4):
    install_requires.extend(read_requirements('py33'))

tags = read_tags(os.path.join('gisht', '__init__.py'))
__doc__ = __doc__.format(**tags)

setup(
    name='gisht',
    version=tags['version'],
    description=tags['description'],
    long_description=__doc__,
    author=tags['author'],
    url="http://github.com/Xion/gisht",
    license=tags['license'],

    platforms='posix',
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: POSIX",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: System :: Software Distribution",
        "Topic :: Utilities",
    ],

    packages=find_packages(exclude=['tests']),
    zip_safe=False,
    entry_points={
        'console_scripts': ['gisht=gisht.__main__:main'],
    },

    install_requires=install_requires,
    tests_require=read_requirements('test'),
)



"""
Module containing gist operations' code.
"""
from __future__ import print_function, unicode_literals

from collections import OrderedDict
import os
from pathlib import Path
import sys
import webbrowser

import envoy
from tabulate import tabulate

from gisht import BIN_DIR, GISTS_DIR, logger
from gisht.github import get_gist_info, iter_gists
from gisht.util import ensure_path, fatal


__all__ = [
    'run_gist', 'output_gist_binary_path', 'print_gist',
    'open_gist_page', 'show_gist_info',

    'gist_exists', 'download_gist',
]


def run_gist(gist, args=()):
    """Run the gist specified by owner/name string.

    This function does not return, because the whole process
    is replaced by the gist's executable.

    :param args: Arguments to pass to the gist
    """
    logger.info("running gist %s ...", gist)

    # TODO(xion): check for the existence of proper shebang,
    # and if it's not there, deduce correct interpreter based on extension
    # of the symlinks target
    cmd = bytes(BIN_DIR / gist)
    os.execv(cmd, [cmd] + list(args))


def output_gist_binary_path(gist):
    """Print the bath to gist binary."""
    print(BIN_DIR / gist)


def print_gist(gist):
    """Print the source code of the gist specified by owner/name string."""
    logger.debug("showing source code for gist %s", gist)

    # resolve the gist exec symlink to find the source file
    # TODO(xion): what about other possible files?
    gist_exec = (BIN_DIR / gist).resolve()
    if gist_exec.exists():
        logger.debug("executable for gist %s found at %s", gist, gist_exec)
    else:
        # TODO(xion): inconsistent state; we should detect those and clean up
        fatal("executable for gist %s missing!", gist, exitcode=os.EX_SOFTWARE)

    with gist_exec.open() as f:
        sys.stdout.write(f.read())


def open_gist_page(gist):
    """Open the gist's GitHub page in the default web browser."""
    logger.debug("opening GitHub page for gist %s...", gist)

    gist_id = get_gist_id(gist)
    gist_info = get_gist_info(gist_id)

    url = gist_info.get('html_url')
    if not url:
        fatal("unable to determine the URL of gist %s...", gist)

    webbrowser.open_new_tab(url)


#: Mapping of gist --info labels to keys in the GitHub API response
#: that describes a gist. Used when displaying information abou a gist.
GIST_INFO_FIELDS = OrderedDict([
    ("ID", 'id'),
    ("Owner", ('owner', 'login')),
    ("URL", 'html_url'),  # URL to gist's user-facing page
    ("Description", 'description'),
    ("Files", ('files', list)),
    ("Created at", 'created_at'),
    ("Comments #", 'comments'),
    ("Forks #", ('forks', len)),
    ("Revisions #", ('history', len)),
    ("Last update", 'updated_at'),
])


def show_gist_info(gist):
    """Shows information about the gist specified by owner/name string."""
    logger.debug("fetching information about gist %s ...", gist)

    gist_id = get_gist_id(gist)
    gist_info = get_gist_info(gist_id)
    logger.info('information about gist %s retrieved successfully', gist)

    # prepare the gist information for display
    info = []
    for label, field in GIST_INFO_FIELDS.items():
        if not isinstance(field, tuple):
            field = (field,)
        data = gist_info
        for step in field:
            data = step(data) if callable(step) else data[step]
        if isinstance(data, list):
            data = ", ".join(data)
        info.append((label, data))

    # print it all as a table
    table = tabulate(((label, ": " + str(value)) for label, value in info),
                     tablefmt='plain')
    print(table)


# Downloading & caching

def gist_exists(gist):
    """Checks if the gist specified by owner/name string exists."""
    gist_exec_symlink = BIN_DIR / gist
    return gist_exec_symlink.exists()  # also checks if symlink is not broken


def get_gist_id(gist):
    """Convert the gist specified by owner/name to gist ID."""
    if not gist_exists(gist):
        fatal("unknown gist %s")

    gist_exec = (BIN_DIR / gist).resolve()
    gist_id = gist_exec.parent.name
    logger.debug("gist %s found to have ID=%s", gist, gist_id)

    return gist_id


def download_gist(gist):
    """Download the gist specified by owner/name string.

    :return: Whether the gist has been successfully downloaded
    """
    # TODO(xion): make this idempotent, i.e. allowing gist to be downloaded
    # multiple times if necessary (to perform updates)

    logger.debug("downloading gist %s ...", gist)

    owner, gist_name = gist.split('/', 1)
    for gist_json in iter_gists(owner):
        for filename in gist_json['files'].keys():
            if filename != gist_name:
                continue

            # clone the gist's repository into directory named after gist ID
            logger.debug("gist %s found, cloning its repository...", gist)
            gist_dir = GISTS_DIR / str(gist_json['id'])
            ensure_path(gist_dir)
            git_clone_run = run('git clone %s %s' % (
                gist_json['git_pull_url'], gist_dir))
            if git_clone_run.status_code != 0:
                logger.warning(
                    "cloning repository for gist %s failed (exitcode %s)",
                    gist, git_clone_run.status_code)
                join(git_clone_run)
            logger.debug("gist %s successfully cloned", gist)

            # make sure the gist executable is, in fact, executable
            gist_exec = gist_dir / filename
            gist_exec.chmod(int('755', 8))
            logger.debug("gist file %s made executable", gist_exec)

            # create symlink from BIN_DIR/<owner>/<gist_name>
            # to the gist's executable file
            gist_owner_bin_dir = BIN_DIR / owner
            ensure_path(gist_owner_bin_dir)
            gist_link = gist_owner_bin_dir / filename
            gist_link.symlink_to(path_vector(from_=gist_link, to=gist_exec))
            logger.debug("symlinked gist 'binary' %s to executable %s",
                         gist_link, gist_exec)

            logger.info("gist %s downloaded sucessfully", gist)
            return True

    return False


# Utility functions

def run(cmd, *args, **kwargs):
    """Wrapper around ``envoy.run`` that ensures the passed command string
    is NOT Unicode string, but a plain buffer of bytes.

    This is necessary to fix some Envoy's command parsing malfeasances.
    """
    return envoy.run(bytes(cmd), *args, **kwargs)


def join(process):
    """Join the process, i.e. pipe its output to our own standard stream
    and relay its exit code back to the system.

    :param process: envoy's process result object
    """
    sys.stdout.write(process.std_out)
    sys.stderr.write(process.std_err)
    raise SystemExit(process.status_code)


def path_vector(from_, to):
    """Return a 'path vector' from given path to the other, i.e.
    the argument of ``cd`` that'd take the user from the source
    directly to target.
    """
    from_, to = map(Path, (from_, to))

    # TODO(xion): consider using http://stackoverflow.com/a/21499676/434799
    # instead of standard os.commonprefix (we don't run into the edge case
    # of the latter yet)
    common_prefix = os.path.commonprefix(list(map(str, (from_, to))))

    # compute the number of '..' segments that are necessary to go
    # from source up to the common prefix
    prefix_wrt_source = Path(from_).relative_to(common_prefix)
    pardir_count = len(prefix_wrt_source.parts)
    if not from_.is_dir():
        pardir_count -= 1

    # join those '..' segments with relative path from common prefix to target
    target_wrt_prefix = Path(to).relative_to(common_prefix)
    return Path(*([os.path.pardir] * pardir_count)) / target_wrt_prefix
